/**
  *  \file planetscentral/file.cc
  *  \brief File Server
  */

#include <vector>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>
#include <pthread.h>
#include <string.h>
#include "server/net.h"
#include "cpluslib/string.h"
#include "server/config.h"
#include "server/request.h"
#include "cpluslib/types.h"
#include "planetscentral/userdir.h"
#include "cpluslib/ptrmap.h"
#include "planetscentral/fsout.h"
#include "planetscentral/pathresolver.h"
#include "io/dirs.h"
#include "util/misc.h"
#include "util/modinit.h"
#include "arch/platform.h"
#include "game/gamedirs.h"
#include "cpluslib/misc.h"

static string_t arg_listen_host = "127.0.0.1";
static string_t arg_listen_port = "9998";
static string_t arg_basedir = ".";
static int32_t arg_file_size_limit = 10*1024*1024;
static int32_t arg_threads = 5;

namespace {
    class MutexGuard {
        pthread_mutex_t* mtx;
     public:
        MutexGuard(pthread_mutex_t* mtx)
            : mtx(mtx)
            { pthread_mutex_lock(mtx); }
        ~MutexGuard()
            { pthread_mutex_unlock(mtx); }
    };
    class Connection {
        Request req;
        int fd;
        pthread_mutex_t* mtx;
        UserRoot& root;
        string_t user;
        bool handleRequest();
        void doError(int n);
        void doStat(FileServerOutput& out, string_t name);
        void doStat(FileServerOutput& out, UserItem* it);
        void doList(FileServerOutput& out, string_t name);
        void doMkdir(string_t name, string_t forUser);
        void doMkdirHier(string_t name);
        void doGet(string_t name);
        void doPut(string_t name, const string_t& data);
        void doCopy(string_t from, string_t to);
        void doRemove(string_t name);
        void doRmdir(string_t name);
        void doForget(string_t name);
        void doUsage(string_t name);
        void doStatReg(FileServerOutput& out, string_t name, DirInfo& info);
        void doStatReg(string_t name);
        void doListReg(string_t name);
        void doStatGame(FileServerOutput& out, string_t name, UserDirectory& dir);
        void doStatGame(string_t name);
        void doListGame(string_t name);
        void doPropGet(string_t name, string_t prop);
        void doPropSet(string_t name, string_t prop, string_t val);
        void doSetPerms(string_t name, string_t user, string_t perms);
        void doListPerms(string_t name);
        void doFileTest(const std::vector<string_t>& names, uint32_t startAt);
     public:
        Connection(int fd, pthread_mutex_t* mtx, UserRoot& root)
            : fd(fd),
              mtx(mtx),
              root(root)
            { }
        void handleConnection();
    };
}

static bool
isPrintable(const string_t& s)
{
    for (string_t::size_type i = 0; i < s.size(); ++i)
        if (!isalnum(uint8_t(s[i])) && s[i] != '/' && s[i] != '.' && s[i] != '_' && s[i] != '-' && s[i] != '*')
            return false;
    return true;
}

/** Read wrapped string. Processes all possible formats:
    - "+XXX" or ":XXX" for single-line strings
    - "$n" + data for multi-line, binary data
    \param req [in] Request (buffering) object
    \param fd  [in] Socket
    \param out [out] Output produced here
    \return true on success, false on syntax error */
static bool
readWrappedString(Request& req, int fd, string_t& out)
{
    /* Read a line */
    const char* command;
    while ((command = req.readLine()) == 0) {
        if (req.readMore(fd) != Request::sSuccess && !req.shift()) {
            return false;
        }
    }

    if (*command == '+' || *command == ':') {
        /* in-line string */
        out += (command+1);
        return true;
    } else if (*command == '$') {
        /* counted string */
        // FIXME: must ensure maximum (file) size here!
        char* end;
        uint32_t n = strtoul(command+1, &end, 10);
        if (*end != 0) {
            return false;
        }

        /* Prepare string */
        string_t::size_type outptr = out.size();
        out.resize(outptr + n);

        /* Read buffered data */
        uint32_t buffered = req.readData(&out[outptr], n);
        outptr += buffered;
        n -= buffered;

        /* Read directly */
        while (n > 0) {
            int direct = read(fd, &out[outptr], n);
            if (direct < 0) {
                return false;
            }
            if (direct == 0) {
                break;
            }
            outptr += direct;
            n -= direct;
        }

        /* Truncate string */
        if (out.size() != outptr)
            out.resize(outptr);

        /* Read a (blank) line */
        while ((command = req.readLine()) == 0) {
            if (req.readMore(fd) != Request::sSuccess && !req.shift()) {
                return false;
            }
        }
        return true;
    } else {
        /* error */
        return false;
    }
}

/** Handle single request. Reads a command and arguments, and
    processes them.
    \retval true Continue processing requests
    \retval false Stop processing requests and close connection */
bool
Connection::handleRequest()
{
    /* Read command */
    char* command;
    while ((command = req.readLine()) == 0) {
        Request::State st = req.readMore(fd);
        if (st == Request::sEOF) {
            return false;
        }
        if (st == Request::sTooLong && req.shift()) {
            /* ok, try again */
        } else if (st != Request::sSuccess) {
            writeStr(fd, "-400 Bad request\r\n");
            return true;
        }
    }

    /* What kind of command is it? */
    std::vector<string_t> data;
    if (*command == '*') {
        /* redis protocol */
        char* end;
        long n = strtol(command+1, &end, 10);
        if (n <= 0 || n > 100 || *end != 0) {
            printf("c2file: bad wrapped command with argc='%s'\n", command);
            writeStr(fd, "-400 Bad request\r\n");
            return true;
        }
        while (n-- > 0) {
            data.push_back(string_t());
            if (!readWrappedString(req, fd, data.back())) {
                printf("c2file: bad wrapped command, truncated/damaged\n");
                writeStr(fd, "-400 Bad request\r\n");
                return true;
            }
        }
    } else {
        /* simple protocol: just split into words */
        while (*command) {
            while (*command == ' ')
                ++command;
            char* p = command;
            while (*p != ' ' && *p != 0)
                ++p;
            if (p != command) {
                data.push_back(string_t(command, p - command));
                command = p;
            }
        }
    }

    /* Anything here? */
    if (data.size() == 0) {
        writeStr(fd, "-400 Bad request\r\n");
        return true;
    }

    /* Upcase first word */
    for (string_t::size_type i = 0; i < data[0].size(); ++i) {
        data[0][i] = toupper(uint8_t(data[0][i]));
    }

    /* Make sure we're the only one from now on */
    MutexGuard guard(mtx);

    /* Log it */
    if (user.size()) {
        printf("[%s]", user.c_str());
    }
    for (uint32_t i = 0; i < data.size(); ++i) {
        printf("%s%s",
               i==0 ? "> " : " ",
               (data[i].size() < 100 && isPrintable(data[i])) ? data[i].c_str() : "...");
    }
    printf("\n");

    /* Process it */
    if (data[0] == "QUIT") {
        /* @q QUIT (File Command)
           Close connection. */
        return false;
    } else if (data[0] == "HELP") {
        /* @q HELP (File Command)
           @retval Str Help page. */
        FileServerOutput(fd).writeFullString("List of commands:\n"
                                             "QUIT\n"
                                             "HELP\n"
                                             "PING\n"
                                             "STAT file\n"
                                             "LS dir\n"
                                             "USER user\n"
                                             "MKDIR dir\n"
                                             "MKDIRAS dir user\n"
                                             "GET file\n"
                                             "PUT file data\n"
                                             "CP from to\n"
                                             "RM file-or-dir\n"
                                             "RMDIR dir\n"
                                             "FORGET dir\n"
                                             "USAGE dir\n"
                                             "STATREG dir\n"
                                             "LSREG dir\n"
                                             "STATGAME dir\n"
                                             "LSGAME dir\n"
                                             "PROPGET dir prop\n"
                                             "PROPSET dir prop value\n"
                                             "SETPERM dir user perms\n"
                                             "LSPERM dir\n"
                                             "FTEST file...\n");
        return true;
    } else if (data[0] == "PING") {
        /* @q PING (File Command)
           Alive test.
           @retval Str "PONG". */
        writeStr(fd, "+PONG\r\n");
        return true;
    } else if (data[0] == "STAT") {
        /* @q STAT file:FileName (File Command)
           Get file status.
           @err 404 Not found (file does not exist)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @retval FileStat file information */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            FileServerOutput out(fd);
            doStat(out, data[1]);
        }
        return true;
    } else if (data[0] == "LS") {
        /* @q LS dir:FileName (File Command)
           List directory content.
           @err 404 Not found (file does not exist)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @retval Hash keys are file/folder names, content is {@type FileStat} elements.
           @rettype FileStat */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            FileServerOutput out(fd);
            doList(out, data[1]);
        }
        return true;
    } else if (data[0] == "USER") {
        /* @q USER user:UID (File Command)
           Set context (caller) for following commands on this connection. */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            user = data[1];
        }
        writeStr(fd, "+OK\r\n");
        return true;
    } else if (data[0] == "MKDIR") {
        /* @q MKDIR dir:FileName (File Command)
           Create directory.
           The containing directory must already exist.
           @err 404 Not found (file does not exist)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @err 409 Exists (directory already exists)
           @err 450 Failed (operating system failure) */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            doMkdir(data[1], string_t());
        }
        return true;
    } else if (data[0] == "MKDIRHIER") {
        /* @q MKDIRHIER dir:FileName (File Command)
           Create directory hierarchy.
           Like {MKDIR}, but can create multiple levels at once,
           and does not fail if one component already exists.
           @err 404 Not found (file does not exist)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @err 409 Exists (name already exists but is not a directory)
           @err 450 Failed (operating system failure) */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            doMkdirHier(data[1]);
        }
        return true;
    } else if (data[0] == "MKDIRAS") {
        /* @q MKDIRAS dir:FileName, user:UID (File Command)
           Create a directory as user.
           The user will become owner of the new directory.
           This is used when creating new user accounts.
           The %user must not be an empty string.
           @err 404 Not found (file does not exist)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @err 409 Exists (directory already exists)
           @err 450 Failed (operating system failure) */
        if (data.size() != 3) {
            writeStr(fd, "-400 Need two args\r\n");
        } else if (data[2].empty()) {
            doError(400);
        } else {
            doMkdir(data[1], data[2]);
        }
        return true;
    } else if (data[0] == "GET") {
        /* @q GET file:FileName (File Command)
           Get file content.
           @err 404 Not found (file does not exist)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @err 500 Internal error (operating system error)
           @err 413 File too large (file cannot be passed through file service; see {File.SizeLimit (Config)})
           @retval Blob file content */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            doGet(data[1]);
        }
        return true;
    } else if (data[0] == "PUT") {
        /* @q PUT file:Str, content:Blob (File Command)
           Create file.
           @err 404 Not found (file does not exist)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @err 450 Unable to create file (operating system error) */
        if (data.size() != 3) {
            writeStr(fd, "-400 Need two args\r\n");
        } else {
            doPut(data[1], data[2]);
        }
        return true;
    } else if (data[0] == "CP") {
        /* @q CP from:FileName, to:FileName (File Command)
           Copies a file.
           @err 404 Not found (file does not exist)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @err 450 Unable to create file (operating system error)
           @err 500 Internal error (operating system error)
           @err 413 File too large (file cannot be passed through file service; see {File.SizeLimit (Config)}) */
        if (data.size() != 3) {
            writeStr(fd, "-400 Need two args\r\n");
        } else {
            doCopy(data[1], data[2]);
        }
        return true;
    } else if (data[0] == "RM") {
        /* @q RM dir:FileName (File Command)
           Remove file or directory.
           A directory must be empty to be removable using RM.
           @err 404 Not found (file does not exist)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @see RMDIR */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            doRemove(data[1]);
        }
        return true;
    } else if (data[0] == "RMDIR") {
        /* @q RMDIR dir:FileName (File Command)
           Remove directory, recursively.
           @err 404 Not found (file does not exist)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @see RM */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            doRmdir(data[1]);
        }
        return true;
    } else if (data[0] == "FORGET") {
        /* @q FORGET dir:FileName (File Command)
           Forget cached information about a directory.
           This is used to tell the file server about an external change to the underlying files.
           This command cannot fail. */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            doForget(data[1]);
        }
        return true;
    } else if (data[0] == "USAGE") {
        /* @q USAGE dir:FileName (File Command)
           Get directory statistics
           @err 404 Not found (file does not exist)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @retkey files:Int Number of files/directories
           @retkey kbytes:Int Disk usage in kilobytes */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            doUsage(data[1]);
        }
        return true;
    } else if (data[0] == "STATREG") {
        /* @q STATREG dir:FileName (File Command)
           Get registration information
           @err 404 Not found (directory does not exist or no registration key)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @retval FileRegInfo Information about this directory */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            doStatReg(data[1]);
        }
        return true;
    } else if (data[0] == "LSREG") {
        /* @q LSREG dir:FileName (File Command)
           List registrations, recursively.
           @err 404 Not found (directory does not exist)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @retval FileRegInfo[] All registrations in this directory and its subdirectories. */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            doListReg(data[1]);
        }
        return true;
    } else if (data[0] == "STATGAME") {
        /* @q STATGAME dir:FileName (File Command)
           Game information for one directory.
           @err 404 Not found (directory does not exist or does not contain a game)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @retval FileGameInfo Information about this directory */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            doStatGame(data[1]);
        }
        return true;
    } else if (data[0] == "LSGAME") {
        /* @q LSGAME dir:FileName (File Command)
           Game information, recursively
           @err 404 Not found (directory does not exist game)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @retval FileGameInfo[] All games in this directory and its subdirectories */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            doListGame(data[1]);
        }
        return true;
    } else if (data[0] == "PROPGET") {
        /* @q PROPGET dir:FileName, prop:Str (File Command)
           Get directory property.
           @err 404 Not found (directory does not exist game)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @retval Str property value
           @see PROPSET */
        if (data.size() != 3) {
            writeStr(fd, "-400 Need two args\r\n");
        } else {
            doPropGet(data[1], data[2]);
        }
        return true;
    } else if (data[0] == "PROPSET") {
        /* @q PROPSET dir:FileName, prop:Str, value:Str (File Command)
           Set directory property.
           A directory can have an arbitrary number of properties attached.
           These store "invisible" meta-information.

           Property names can contain any character except "\r", "\n" and "=".
           Property values can contain any character except "\r", "\n".

           @err 404 Not found (directory does not exist game)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name, property name, or value) */
        if (data.size() != 4) {
            writeStr(fd, "-400 Need three args\r\n");
        } else {
            doPropSet(data[1], data[2], data[3]);
        }
        return true;
    } else if (data[0] == "SETPERM") {
        /* @q SETPERM dir:FileName, user:UID, perms:FilePermissions (File Command)
           Set directory permissions.
           The owner of a directory can grant permissions for other users to access their files.
           The special %user value "*" defines permissions for everyone (world permissions).
           @err 404 Not found (directory does not exist game)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name) */
        if (data.size() != 4) {
            writeStr(fd, "-400 Need three args\r\n");
        } else {
            doSetPerms(data[1], data[2], data[3]);
        }
        return true;
    } else if (data[0] == "LSPERM") {
        /* @q LSPERM dir:FileName (File Command)
           List permissions.
           @err 404 Not found (directory does not exist game)
           @err 403 Forbidden (insufficient permissions)
           @err 400 Bad request (invalid file name)
           @retkey owner:UID (owner of the directory)
           @retkey perms:List (list of hashes, each with keys %user:{@type UID} and %perms:{@type FilePermissions})
           @rettype UID
           @rettype FilePermissions */
        if (data.size() != 2) {
            writeStr(fd, "-400 Need one arg\r\n");
        } else {
            doListPerms(data[1]);
        }
        return true;
    } else if (data[0] == "FTEST") {
        /* @q FTEST file:FileName... (File Command)
           Quick file test.
           This command quickly checks accessability of a bunch of files.
           It returns one integer for each file: 0=does not exist or not readable, 1=exists and readable
           @retval IntList results */
        doFileTest(data, 1);
        return true;
    } else {
        printf("c2file: unknown command '%s'\n", data[0].c_str());
        writeStr(fd, "-405 Unknown command\r\n");
        return true;
    }
}

void
Connection::doError(int n)
{
    if (n == 404)
        writeStr(fd, "-404 Not found\r\n");
    else if (n == 403)
        writeStr(fd, "-403 Forbidden\r\n");
    else
        writeStr(fd, "-400 Bad request\r\n");
}

void
Connection::doStat(FileServerOutput& out, string_t name)
{
    PathResolver res(&root, user);
    UserItem* it = res.resolve(name, 0);
    if (!it) {
        doError(res.getError());
    } else {
        doStat(out, it);
    }
}

void
Connection::doStat(FileServerOutput& out, UserItem* it)
{
    if (UserDirectory* d = dynamic_cast<UserDirectory*>(it)) {
        d->read();
        out.writeHashStart(2);
        out.writeSimpleString("type");
        out.writeSimpleString("dir");
        out.writeSimpleString("visibility");
        out.writeNumber(d->getVisibilityLevel());
    } else if (UserFile* f = dynamic_cast<UserFile*>(it)) {
        out.writeHashStart(2);
        out.writeSimpleString("type");
        out.writeSimpleString("file");
        out.writeSimpleString("size");
        out.writeNumber(f->getSize());
    } else {
        out.writeHashStart(1);
        out.writeSimpleString("type");
        out.writeSimpleString("unknown");
    }
}

void
Connection::doList(FileServerOutput& out, string_t name)
{
    PathResolver res(&root, user);
    UserItem* it = res.resolve(name, 0);
    if (!it) {
        doError(res.getError());
    } else if (UserDirectory* d = dynamic_cast<UserDirectory*>(it)) {
        d->read();
        if (d->getPermissions(user) & UserDirectory::AllowList) {
            out.writeHashStart(d->getNumDirs() + d->getNumFiles());
            for (uint32_t i = 0, n = d->getNumDirs(); i < n; ++i) {
                out.writeFullString(d->getDir(i)->getName());
                doStat(out, d->getDir(i));
            }
            for (uint32_t i = 0, n = d->getNumFiles(); i < n; ++i) {
                out.writeFullString(d->getFile(i)->getName());
                doStat(out, d->getFile(i));
            }
        } else {
            doError(403);
        }
    } else {
        writeStr(fd, "-405 Not a directory\r\n");
    }
}

void
Connection::doMkdir(string_t path, string_t forUser)
{
    /* Resolve path */
    PathResolver res(&root, user);
    if (!res.resolvePath(path)) {
        doError(res.getError());
        return;
    }
    if (!res.checkPrivilege(UserDirectory::AllowWrite)) {
        doError(403);
        return;
    }
    if (res.resolveLeaf(path)) {
        writeStr(fd, "-409 Exists\r\n");
        return;
    }

    if (res.getDirectory()->createDirectory(path, forUser)) {
        writeStr(fd, "+OK\r\n");
    } else {
        writeStr(fd, "-450 Failed\r\n");
    }
}

void
Connection::doMkdirHier(string_t path)
{
    /* Simple dumb-ass version: try to create path */
    for (string_t::size_type i = 0; i <= path.size(); ++i) {
        if (i == path.size() || path[i] == '/') {
            /* Got a partial path name */
            string_t part(path, 0, i);
            PathResolver res(&root, user);
            if (!res.resolvePath(part)) {
                /* Part cannot be resolved */
                doError(res.getError());
                return;
            }
            UserItem* it = res.resolveLeaf(part);
            if (it != 0) {
                if (dynamic_cast<UserDirectory*>(it) == 0) {
                    /* Item exists but is not a directory */
                    writeStr(fd, "-409 Exists\r\n");
                    return;
                }
            } else {
                if (!res.checkPrivilege(UserDirectory::AllowWrite)) {
                    /* We'll be creating something, so we need write permission */
                    doError(403);
                    return;
                }
                if (!res.getDirectory()->createDirectory(part, string_t())) {
                    writeStr(fd, "-450 Failed\r\n");
                    return;
                }
            }
        }
    }
    writeStr(fd, "+OK\r\n");
}

void
Connection::doGet(string_t name)
{
    PathResolver res(&root, user);
    if (!res.resolvePath(name)) {
        doError(res.getError());
        return;
    }
    if (!res.checkPrivilege(UserDirectory::AllowRead)) {
        doError(403);
        return;
    }
    if (!dynamic_cast<UserFile*>(res.resolveLeaf(name))) {
        doError(404);
        return;
    }

    Ptr<Stream> s = res.getDirectory()->openFile(name, Stream::C_READ);
    if (!s) {
        writeStr(fd, "-500 Internal error\r\n");
        return;
    }

    int32_t size = s->getSize();
    if (size < 0 || size > arg_file_size_limit) {
        writeStr(fd, "-413 File too large\r\n");
        return;
    }

    string_t buffer;
    buffer.resize(size);
    s->read(&buffer[0], size);
    FileServerOutput(fd).writeFullString(buffer);
}

void
Connection::doPut(string_t name, const string_t& data)
{
    PathResolver res(&root, user);
    if (!res.resolvePath(name)) {
        doError(res.getError());
        return;
    }
    if (!res.checkPrivilege(UserDirectory::AllowWrite)) {
        doError(403);
        return;
    }
    if (!res.getDirectory()->createFile(name, data)) {
        writeStr(fd, "-450 Unable to create file\r\n");
    } else {
        writeStr(fd, "+OK\r\n");
    }
}

void
Connection::doCopy(string_t from, string_t to)
{
    /* Resolve from [see GET] */
    PathResolver res_from(&root, user);
    if (!res_from.resolvePath(from)) {
        doError(res_from.getError());
        return;
    }
    if (!res_from.checkPrivilege(UserDirectory::AllowRead)) {
        doError(403);
        return;
    }
    if (!dynamic_cast<UserFile*>(res_from.resolveLeaf(from))) {
        doError(404);
        return;
    }

    /* Resolve to [see PUT] */
    PathResolver res_to(&root, user);
    if (!res_to.resolvePath(to)) {
        doError(res_to.getError());
        return;
    }
    if (!res_to.checkPrivilege(UserDirectory::AllowWrite)) {
        doError(403);
        return;
    }

    /* Read */
    string_t buffer;
    {
        Ptr<Stream> s = res_from.getDirectory()->openFile(from, Stream::C_READ);
        if (!s) {
            writeStr(fd, "-500 Internal error\r\n");
            return;
        }

        int32_t size = s->getSize();
        if (size < 0 || size > arg_file_size_limit) {
            writeStr(fd, "-413 File too large\r\n");
            return;
        }

        buffer.resize(size);
        s->read(&buffer[0], size);
    }

    /* Write */
    if (!res_to.getDirectory()->createFile(to, buffer)) {
        writeStr(fd, "-450 Unable to create file\r\n");
    } else {
        writeStr(fd, "+OK\r\n");
    }
}

void
Connection::doRemove(string_t name)
{
    /* Resolve path. We need write access to name's parent. */
    PathResolver res(&root, user);
    if (!res.resolvePath(name)) {
        doError(res.getError());
        return;
    }
    if (!res.checkPrivilege(UserDirectory::AllowWrite)) {
        doError(403);
        return;
    }

    /* Resolve final component */
    UserItem* it = res.resolveLeaf(name);
    if (!it) {
        doError(404);
    } else {
        if (!res.getDirectory()->removeEntry(it)) {
            doError(403);
        } else {
            writeStr(fd, "+OK\r\n");
        }
    }
}

void
Connection::doRmdir(string_t name)
{
    /* Resolve path. We need write access to name's parent. */
    PathResolver res(&root, user);
    if (!res.resolvePath(name)) {
        doError(res.getError());
        return;
    }
    if (!res.checkPrivilege(UserDirectory::AllowWrite)) {
        doError(403);
        return;
    }

    /* Resolve final component */
    UserDirectory* dir = dynamic_cast<UserDirectory*>(res.resolveLeaf(name));
    if (!dir) {
        doError(404);
        return;
    }

    /* Fully read it */
    dir->readFully();

    /* Build list of directories. While doing so, check permissions.
       Must have write permission for everyone. */
    std::vector<UserDirectory*> dirs;
    dirs.push_back(dir);
    uint32_t index = 0;
    while (index < dirs.size()) {
        /* Fetch directory */
        UserDirectory* it = dirs[index++];
        if ((it->getPermissions(user) & UserDirectory::AllowWrite) == 0) {
            doError(403);
            return;
        }

        /* Enumerate subdirectories */
        for (uint32_t i = 0, e = it->getNumDirs(); i < e; ++i) {
            dirs.push_back(it->getDir(i));
        }
    }

    /* We now have all directories in dirs, parents preceding
       their children. We can therefore delete them recursively
       by just going backwards. */
    for (uint32_t i = dirs.size(); i > 0; --i) {
        if (!dirs[i-1]->removeAllEntries()) {
            doError(403);
            return;
        }
    }

    /* Finally, delete the directory itself */
    if (!res.getDirectory()->removeEntry(dir)) {
        doError(403);
    } else {
        writeStr(fd, "+OK\r\n");
    }
}

void
Connection::doForget(string_t name)
{
    /* We want to avoid pulling in directories that do not exist or
       are not loaded, and we do not need permission or syntax checks
       because we're not giving out information to the user. So we
       implement our own, simple path parser here. */
    UserDirectory* p = &root;
    string_t::size_type n;
    while (p != 0 && (n = name.find('/')) != string_t::npos) {
        p = p->wasRead() ? dynamic_cast<UserDirectory*>(p->find(string_t(name, 0, n))) : 0;
        name.erase(0, n+1);
    }
    if (p != 0) {
        p = p->wasRead() ? dynamic_cast<UserDirectory*>(p->find(name)) : 0;
    }
    if (p != 0) {
        p->forget();
    }
    writeStr(fd, "+OK\r\n");
}

void
Connection::doUsage(string_t name)
{
    /* Resolve path */
    PathResolver res(&root, user);
    if (!res.resolvePath(name)) {
        doError(res.getError());
        return;
    }
    UserItem* it = res.resolveLeaf(name);
    if (!it) {
        doError(res.getError());
        return;
    }
    UserDirectory* dir = dynamic_cast<UserDirectory*>(it);
    if (!dir) {
        doError(404);
        return;
    }

    /* Get stats */
    int32_t nfiles = 0, nkbytes = 0;
    dir->readFully();
    dir->computeTotals(nfiles, nkbytes);

    FileServerOutput out(fd);
    out.writeHashStart(2);
    out.writeSimpleString("files");
    out.writeNumber(nfiles);
    out.writeSimpleString("kbytes");
    out.writeNumber(nkbytes);
}

void
Connection::doStatReg(FileServerOutput& out, string_t name, DirInfo& info)
{
    out.writeHashStart(5);
    out.writeSimpleString("path");
    out.writeFullString(name);
    out.writeSimpleString("file");
    out.writeFullString(name + "/" + info.getRegFileName());
    out.writeSimpleString("reg");
    out.writeNumber(info.getRegInfo().isRegistered());
    out.writeSimpleString("key1");
    out.writeFullString(info.getRegInfo().getRegStr1());
    out.writeSimpleString("key2");
    out.writeFullString(info.getRegInfo().getRegStr2());
}

void
Connection::doStatReg(string_t name)
{
    /* Resolve path */
    PathResolver res(&root, user);
    UserDirectory* dir = res.resolveToDirectory(name, UserDirectory::AllowRead);
    if (!dir) {
        doError(res.getError());
        return;
    }
    dir->read();

    /* Get directory info */
    DirInfo& info = dir->getDirInfo();
    if (!info.haveRegInfo()) {
        doError(404);
    } else {
        FileServerOutput out(fd);
        doStatReg(out, name, info);
    }
}

void
Connection::doListReg(string_t name)
{
    /* Resolve path */
    PathResolver res(&root, user);
    UserDirectory* dir = res.resolveToDirectory(name, UserDirectory::AllowRead);
    if (!dir) {
        doError(res.getError());
        return;
    }

    /* Process recursively */
    std::vector<std::pair<string_t, UserDirectory*> > work;
    std::vector<std::pair<string_t, DirInfo*> > result;
    work.push_back(std::make_pair(name, dir));
    while (!work.empty()) {
        /* Fetch work item */
        UserDirectory* thisdir = work.back().second;
        string_t thisname = work.back().first;
        work.pop_back();
        thisdir->readFully();

        /* Any info here? */
        DirInfo& info = thisdir->getDirInfo();
        if (info.haveRegInfo()) {
            result.push_back(std::make_pair(thisname, &info));
        }

        /* Gather subdirs */
        for (uint32_t i = 0, n = thisdir->getNumDirs(); i < n; ++i) {
            UserDirectory* subdir = thisdir->getDir(i);
            subdir->read();
            if (subdir->getPermissions(user) & UserDirectory::AllowRead) {
                work.push_back(std::make_pair(thisname + "/" + subdir->getName(), subdir));
            }
        }
    }

    /* Generate result */
    FileServerOutput out(fd);
    out.writeArrayStart(result.size());
    for (uint32_t i = 0; i < result.size(); ++i) {
        doStatReg(out, result[i].first, *result[i].second);
    }
}

void
Connection::doStatGame(FileServerOutput& out, string_t name, UserDirectory& dir)
{
    static const char*const props[] = { "name", "game", "finished", "hosttime" };

    out.writeHashStart(4 + countof(props));
    out.writeSimpleString("path");
    out.writeFullString(name);
    for (uint_t i = 0; i < countof(props); ++i) {
        out.writeSimpleString(props[i]);
        out.writeFullString(dir.getProperty(string_t("prop:") + props[i]));
    }

    DirInfo& info = dir.getDirInfo();

    const GDirectoryOverview& ov = info.getOverview();
    int nraces = 0;
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (ov.getPlayerFlags(i) & ov.have_RST) {
            ++nraces;
        }
    }
    out.writeSimpleString("races");
    out.writeArrayStart(2*nraces);
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (ov.getPlayerFlags(i) & ov.have_RST) {
            out.writeNumber(i);
            if (info.haveRaceNames())
                out.writeFullString(info.getRaceNames().getFullName(i));
            else
                out.writeFullString(host_racenames.getFullName(i));
        }
    }

    out.writeSimpleString("missing");
    const DirInfo::MissingList_t& miss = info.getMissingFiles();
    out.writeArrayStart(miss.size());
    for (DirInfo::MissingList_t::const_iterator i = miss.begin(); i != miss.end(); ++i)
        out.writeFullString(*i);

    out.writeSimpleString("conflict");
    out.writeArrayStart(0);
}

void
Connection::doStatGame(string_t name)
{
    /* Resolve path */
    PathResolver res(&root, user);
    UserDirectory* dir = res.resolveToDirectory(name, UserDirectory::AllowRead);
    if (!dir) {
        doError(res.getError());
        return;
    }
    dir->read();

    /* Get directory info */
    DirInfo& info = dir->getDirInfo();
    if (!info.haveGame()) {
        doError(404);
    } else {
        FileServerOutput out(fd);
        doStatGame(out, name, *dir);
    }
}

void
Connection::doListGame(string_t name)
{
    /* Resolve path */
    PathResolver res(&root, user);
    UserDirectory* dir = res.resolveToDirectory(name, UserDirectory::AllowRead);
    if (!dir) {
        doError(res.getError());
        return;
    }

    /* Process recursively */
    std::vector<std::pair<string_t, UserDirectory*> > work;
    std::vector<std::pair<string_t, UserDirectory*> > result;
    work.push_back(std::make_pair(name, dir));
    while (!work.empty()) {
        /* Fetch work item */
        UserDirectory* thisdir = work.back().second;
        string_t thisname = work.back().first;
        work.pop_back();
        thisdir->readFully();

        /* Any info here? */
        DirInfo& info = thisdir->getDirInfo();
        if (info.haveGame()) {
            result.push_back(std::make_pair(thisname, thisdir));
        }

        /* Gather subdirs */
        for (uint32_t i = 0, n = thisdir->getNumDirs(); i < n; ++i) {
            UserDirectory* subdir = thisdir->getDir(i);
            subdir->read();
            if (subdir->getPermissions(user) & UserDirectory::AllowRead) {
                work.push_back(std::make_pair(thisname + "/" + subdir->getName(), subdir));
            }
        }
    }

    /* Generate result */
    FileServerOutput out(fd);
    out.writeArrayStart(result.size());
    for (uint32_t i = 0; i < result.size(); ++i) {
        doStatGame(out, result[i].first, *result[i].second);
    }
}

void
Connection::doPropGet(string_t name, string_t prop)
{
    PathResolver res(&root, user);
    UserDirectory* dir = res.resolveToDirectory(name, UserDirectory::AllowRead);
    if (!dir) {
        doError(res.getError());
        return;
    }
    dir->read();

    FileServerOutput(fd).writeFullString(dir->getProperty("prop:" + prop));
}

void
Connection::doPropSet(string_t name, string_t prop, string_t val)
{
    PathResolver res(&root, user);
    UserDirectory* dir = res.resolveToDirectory(name, UserDirectory::AllowWrite);
    if (!dir) {
        doError(res.getError());
        return;
    }
    if (val.find_first_of("\r\n") != val.npos || prop.find_first_of("\r\n=") != prop.npos) {
        doError(400);
        return;
    }

    dir->read();
    dir->setProperty("prop:" + prop, val);
    writeStr(fd, "+OK\r\n");
}

void
Connection::doSetPerms(string_t name, string_t forUser, string_t perms)
{
    PathResolver res(&root, user);
    UserDirectory* dir = res.resolveToDirectory(name, UserDirectory::AllowAccess);
    if (!dir) {
        doError(res.getError());
        return;
    }
    dir->read();
    dir->setPermissions(forUser, perms);
    writeStr(fd, "+OK\r\n");
}

void
Connection::doListPerms(string_t name)
{
    PathResolver res(&root, user);
    UserDirectory* dir = res.resolveToDirectory(name, UserDirectory::AllowAccess);
    if (!dir) {
        doError(res.getError());
        return;
    }
    dir->read();

    std::vector<std::pair<string_t, string_t> > perms;
    dir->listPermissions(perms);
    FileServerOutput out(fd);
    out.writeHashStart(2);
    out.writeSimpleString("owner");
    out.writeFullString(dir->getOwner());
    out.writeSimpleString("perms");
    out.writeArrayStart(perms.size());
    for (uint32_t i = 0; i < perms.size(); ++i) {
        out.writeHashStart(2);
        out.writeSimpleString("user");
        out.writeFullString(perms[i].first);
        out.writeSimpleString("perms");
        out.writeFullString(perms[i].second);
    }
}

void
Connection::doFileTest(const std::vector<string_t>& names, uint32_t startAt)
{
    FileServerOutput out(fd);
    out.writeArrayStart(names.size() - startAt);
    while (startAt < names.size()) {
        int ok;
        string_t name = names[startAt];
        PathResolver res(&root, user);
        if (!res.resolvePath(name)) {
            ok = 0;
        } else if (!res.checkPrivilege(UserDirectory::AllowRead)) {
            ok = 0;
        } else if (!dynamic_cast<UserFile*>(res.resolveLeaf(name))) {
            ok = 0;
        } else {
            ok = 1;
        }
        out.writeNumber(ok);
        ++startAt;
    }
}

void
Connection::handleConnection()
{
    while (handleRequest())
        ;
}

static string_t arg_instance("FILE");

static void
processConfig(string_t key, string_t value)
{
    if (key == arg_instance + ".HOST") {
        /* @q File.Host:Str (Config), HostFile.Host:Str (Config)
           Listen address for the File/HostFile instance. */
        arg_listen_host = value;
    } else if (key == arg_instance + ".PORT") {
        /* @q File.Port:Int (Config), HostFile.Port:Int (Config)
           Port number for the File/HostFile instance. */
        arg_listen_port = value;
    } else if (key == arg_instance + ".BASEDIR") {
        /* @q File.BaseDir:Str (Config), HostFile.BaseDir:Str (Config)
           Base directory where managed files are. */
        arg_basedir = value;
    } else if (key == arg_instance + ".SIZELIMIT") {
        /* @q File.SizeLimit:Int (Config), HostFile.SizeLimit:Int (Config)
           Maximum size of a file in this instance. */
        stringValue(value, arg_file_size_limit);
    } else if (key == arg_instance + ".THREADS") {
        /* @q File.Threads:Int (Config), HostFile.Threads:Int (Config)
           Number of threads (=maximum number of parallel connections) */
        stringValue(value, arg_threads);
    }
}

static int server_socket;
static pthread_mutex_t server_mtx;
static UserRoot* server_root;

static void*
serverLoop(void*)
{
    while (1) {
        int fd = accept(server_socket, 0, 0);
        if (fd < 0) {
            perror("accept");
            sleep(10);
        } else {
            fcntl(fd, F_SETFD, FD_CLOEXEC);
            Connection(fd, &server_mtx, *server_root).handleConnection();
            close(fd);
        }
    }
}


static const ModuleInitializer*const file_modules[] = {
    mod_reg,
    mod_racenames,
    mod_base_console,
    0
};

int main(int argc, char** argv)
{
    /* Parse args */
    while (const char* p = *++argv) {
        if (strncmp(p, "-I", 2) == 0) {
            arg_instance = strUCase(p+2);
        } else {
            fprintf(stderr, "unexpected argument: %s\n", p);
            return 1;
        }
    }

    /* Unbuffer stdout */
    setvbuf(stdout, 0, _IOLBF, 0);

    /* Configuration */
    loadConfiguration(processConfig);

    /* Initialize. This will load the default race names. */
    initDirectories(Platform::getDataDirectory(), Platform::getDataDirectory());
    performInitialisation(OnStartup | OnEnterNoGame, file_modules);

    /* Listen */
    server_socket = Bind().withAddress(arg_listen_host, arg_listen_port);
    if (server_socket < 0) {
        fprintf(stderr, "unable to bind socket\n");
        return 1;
    }

    /* Main data structure */
    static UserRoot root("(root)", arg_basedir);
    server_root = &root;

    /* Multithreading */
    if (pthread_mutex_init(&server_mtx, 0) < 0) {
        perror("unable to create mutex");
        return 1;
    }

    /* OK, main loop */
    printf("c2file: listening...\n");
    signal(SIGPIPE, SIG_IGN);
    fcntl(server_socket, F_SETFD, FD_CLOEXEC);
    listen(server_socket, 10);

    /* Create some threads. Create one less than configured; the main
       thread will also serve requests. */
    for (int32_t i = 1; i < arg_threads; ++i) {
        pthread_t handle;
        if (pthread_create(&handle, 0, serverLoop, 0) < 0) {
            perror("unable to create thread");
            return 1;
        }
    }
    serverLoop(0);
    close(server_socket);
}
