/**
  *  \file planetscentral/host/game.cc
  *  \brief Host: Game operations
  */

#include <stdexcept>
#include "planetscentral/host/game.h"
#include "cpluslib/format.h"
#include "game/struct.h"
#include "int/array.h"
#include "io/io.h"
#include "io/memstream.h"
#include "planetscentral/dbclient/fileclient.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/host/cmdsched.h"
#include "planetscentral/host/cron.h"
#include "planetscentral/host/db.h"
#include "planetscentral/host/host.h"
#include "planetscentral/host/schedule.h"
#include "util/misc.h"
#include "planetscentral/host/install.h"
#include "planetscentral/host/rating.h"
#include "planetscentral/host/talk.h"

/******************************* Game::Slot ******************************/

Game::Slot::Slot(DbSubtree tree)
    : tree(tree)
{ }

DbStringList
Game::Slot::players()
{
    return tree.stringListKey("users");
}

DbHashIntField
Game::Slot::slotStatus()
{
    return tree.hashKey("status").intField("slot");
}

DbHashIntField
Game::Slot::turnStatus()
{
    return tree.hashKey("status").intField("turn");
}

DbHashIntField
Game::Slot::rank()
{
    return tree.hashKey("status").intField("rank");
}

/********************************** Game *********************************/

/** Constructor.
    \param root Root of the "game:" database
    \param gid  Game Id
    \throw exception if game does not exist */
Game::Game(DbSubtree root, int32_t gid)
    : root(root),
      game(root.subtree(gid)),
      gameId(gid)
{
    if (!root.intSetKey("all").contains(gid)) {
        throw std::runtime_error("404 Game does not exist");
    }
}

Game::Game(DbSubtree root, int32_t gid, NoExistanceCheck_t)
    : root(root),
      game(root.subtree(gid)),
      gameId(gid)
{ }


/** Get game Id. */
int32_t
Game::getId() const
{
    return gameId;
}

/** Get game state. Valid states are:
    - preparing
    - joining
    - running
    - finished
    - deleted */
string_t
Game::getState()
{
    return game.stringKey("state").get();
}

/** Set game state. Updates all respective lists.
    \param newState new state. \see getState */
void
Game::setState(string_t newState)
{
    string_t oldState = game.stringKey("state").replaceBy(newState);
    if (oldState != newState) {
        // There was a change
        string_t gameType = getType();

        // Update history
        string_t histMessage = newState;
        if (newState == "finished") {
            // The game was finished. Is there a unique victor?
            string_t victor;
            for (int i = 1; i <= NUM_PLAYERS; ++i) {
                Slot slot(getSlot(i));
                if (slot.slotStatus().get() != 0 && slot.rank().get() == 1) {
                    // This is a rank-1 player
                    string_t player = slot.players()[0];
                    if (!player.empty() && !victor.empty()) {
                        // Ambiguous
                        victor.clear();
                        break;
                    }
                    victor = player;
                }
            }
            if (!victor.empty()) {
                histMessage += ":";
                histMessage += victor;
            }
        }
        addGameHistoryItem("game-state", histMessage, gameType == "public" && newState != "preparing" && newState != "deleted");

        // Update indexes
        root.subtree("state").intSetKey(oldState).moveTo(gameId, root.subtree("state").intSetKey(newState));
        if (gameType == "public") {
            root.subtree("pubstate").intSetKey(oldState).moveTo(gameId, root.subtree("pubstate").intSetKey(newState));
        }
        if (newState == "running") {
            if (getConfigInt("copyOf") != 0) {
                setConfigInt("copyPending", 1);
            }
            cron.handleGameStart(gameId);
        }
        if (newState == "joining" || newState == "running") {
            talkHandleGameStart(*this, gameType);
        }
        if (newState == "finished") {
            talkHandleGameEnd(*this, gameType);
        }

        // Global history entry.
    }
}

/** Get game type. Valid types are:
    - public
    - unlisted
    - private
    - test */
string_t
Game::getType()
{
    return game.stringKey("type").get();
}

/** Set game type. */
void
Game::setType(string_t newType)
{
    // Change type
    string_t oldType = game.stringKey("type").replaceBy(newType);
    if (oldType != newType) {
        // There was a change, and we may have to move it
        string_t state = getState();
        if (oldType == "public") {
            root.subtree("pubstate").intSetKey(state).remove(gameId);
        }
        if (newType == "public") {
            root.subtree("pubstate").intSetKey(state).add(gameId);
        }
        talkHandleGameTypeChange(*this, state, newType);
    }
}

/** Get game name. */
string_t
Game::getName()
{
    return game.stringKey("name").get();
}

/** Set game name. */
void
Game::setName(string_t newName)
{
    game.stringKey("name").set(newName);
    talkHandleGameNameChange(*this, newName);
}

/** Get game owner. */
string_t
Game::getOwner()
{
    return game.stringKey("owner").get();
}

/** Set game owner. Updates all respective lists.
    \param newOwner new state
    \param userRoot root of the "user:" database */
void
Game::setOwner(string_t newOwner, DbSubtree userRoot)
{
    string_t oldOwner = game.stringKey("owner").replaceBy(newOwner);
    if (oldOwner != newOwner) {
        if (!oldOwner.empty()) {
            userRoot.subtree(oldOwner).intSetKey("ownedGames").remove(gameId);
        }
        if (!newOwner.empty()) {
            userRoot.subtree(newOwner).intSetKey("ownedGames").add(gameId);
        }
    }
}

/** Get game configuration property. */
string_t
Game::getConfig(string_t name)
{
    return game.hashKey("settings").getStringField(name);
}

/** Set game configuration property. */
void
Game::setConfig(string_t name, string_t value)
{
    game.hashKey("settings").setField(name, value);
}

/** Get multiple configuration properties.
    \param names List of keys
    \return result returned by database; usually, an IntArray. */
IntValue*
Game::getConfig(const StringList_t& names)
{
    return game.hashKey("settings").getFields(names);
}

/** Set multiple configuration properties.
    \param kv List of keys (odd elements) and values (even elements)  */
void
Game::setConfig(const StringList_t& kv)
{
    game.hashKey("settings").setFields(kv);
}

/** Get game configuration property. */
int32_t
Game::getConfigInt(string_t name)
{
    return game.hashKey("settings").getIntField(name);
}

/** Set game configuration property. */
void
Game::setConfigInt(string_t name, int32_t value)
{
    game.hashKey("settings").setField(name, value);
}

/** Remove game configuration property. */
void
Game::removeConfig(string_t name)
{
    game.hashKey("settings").removeField(name);
}

/** Get game directory. */
string_t
Game::getDirectory()
{
    return game.stringKey("dir").get();
}

/** Add a history item.
    \param id Message Name
    \param args Arguments in addition to game Id */
void
Game::addGameHistoryItem(string_t what, string_t args, bool global)
{
    string_t message = format("%d:%s:%d:%s", getCurrentTime(), what, gameId, args);
    game.stringListKey("history").pushFront(message);
    if (global) {
        DbStringList(game.getClient(), GLOBAL_HISTORY).pushFront(message);
    }
}

/** Get player configuration item. */
string_t
Game::getPlayerConfig(string_t player, string_t name)
{
    return game.subtree("user").hashKey(player).getStringField(name);
}

/** Set player configuration item. */
void
Game::setPlayerConfig(string_t player, string_t name, string_t value)
{
    game.subtree("user").hashKey(player).setField(name, value);
}

/** Get player configuration item. */
int32_t
Game::getPlayerConfigInt(string_t player, string_t name)
{
    return game.subtree("user").hashKey(player).getIntField(name);
}

/** Set player configuration item. */
void
Game::setPlayerConfigInt(string_t player, string_t name, int32_t value)
{
    game.subtree("user").hashKey(player).setField(name, value);
}

/** Get name of score used for referee.
    Note that this can return an emtpy string when called before the first host. */
string_t
Game::getRefereeScoreName()
{
    string_t score = getConfig("endScoreName");
    if (score.empty() && getConfigInt("turn") > 0) {
        // We can find an implicit name only for games that have started!
        if (game.hashKey("scores").hasField("score")) {
            score = "score";
        } else {
            score = "timscore";
        }
    }
    return score;
}

/** Add a history item.
    \param id Message Name
    \param args Arguments in addition to game Id
    \param userRoot root of user DB
    \param player user to give this message too */
void
Game::addUserHistoryItem(string_t what, string_t args, DbSubtree userRoot, string_t player)
{
    string_t message = format("%d:%s:%d:%s", getCurrentTime(), what, gameId, args);
    game.stringListKey("history").pushFront(message);
    userRoot.subtree(player).stringListKey("history").pushFront(message);
}

/** Get slot status. Members can be used to access properties of the slot. */
Game::Slot
Game::getSlot(int32_t slot)
{
    return Slot(game.subtree("player").subtree(slot));
}

/** Check whether a slot is available. */
bool
Game::isSlotInGame(int32_t slot)
{
    // Check slot range first; although it would normally fail at the database anyway
    return slot > 0
        && slot <= NUM_PLAYERS
        && getSlot(slot).slotStatus().get() != 0;
}

/** Check whether a slot is played. */
bool
Game::isSlotPlayed(int32_t slot)
{
    return !getSlot(slot).players().empty();
}

/** Check whether the game has any open slots. */
bool
Game::hasAnyOpenSlot()
{
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (isSlotInGame(i) && !isSlotPlayed(i)) {
            return true;
        }
    }
    return false;
}

/** Add player to a slot.
    Note: this function does not start a game getting full.
    This is done by the caller (PLAYERJOIN command). */
void
Game::pushPlayerSlot(int32_t slot, string_t player, DbSubtree userRoot)
{
    // Add to database
    getSlot(slot).players().pushBack(player);
    ++game.hashKey("users").intField(player);
    ++userRoot.subtree(player).hashKey("games").intField(format("%d", gameId));

    // Grant him access to the game's transfer folder
    string_t gameDir = game.stringKey("dir").get();
    {
        ReconnectingGuard g(file_rc);
        FileClient file(file_connection);
        file.setRootPermissions();
        file.setFilePermissions(gameDir + "/in/new", player, "w");
        file.setFilePermissions(gameDir + "/out/all", player, "rl");
        file.setFilePermissions(gameDir + "/out/" + format("%d", slot), player, "rl");
    }

    // Give him player files
    try {
        installChangedGameFiles(*this, player, slot, true);
    }
    catch (std::exception& e) {
        // FIXME: log it [implemented in c2ng]
    }
}

/** Drop player from a slot. */
string_t
Game::popPlayerSlot(int32_t slot, DbSubtree userRoot)
{
    string_t player = getSlot(slot).players().popBack();
    if (!player.empty()) {
        --game.hashKey("users").intField(player);
        int32_t ref = --userRoot.subtree(player).hashKey("games").intField(format("%d", gameId));

        // Revoke file permissions
        string_t gameDir = game.stringKey("dir").get();
        {
            ReconnectingGuard g(file_rc);
            FileClient file(file_connection);
            file.setRootPermissions();
            file.setFilePermissions(gameDir + "/out/" + format("%d", slot), player, "0");
            if (ref == 0) {
                // Revoke file permissions for public directories
                file.setFilePermissions(gameDir + "/in/new", player, "0");
                file.setFilePermissions(gameDir + "/out/all", player, "0");
            }
        }

        // Uninstall game files
        try {
            installChangedGameFiles(*this, player, slot, false);
        }
        catch (std::exception& e) {
            // FIXME: log it [implemented in c2ng]
        }
    }
    return player;
}

/** List players for a slot. */
void
Game::listPlayers(int32_t slot, StringList_t& players)
{
    getSlot(slot).players().getAll(players);
}

/** Get set of slots played by a player. */
GPlayerSet
Game::getSlotsByPlayer(string_t player)
{
    GPlayerSet slots;
    for (int32_t slot = 1; slot <= NUM_PLAYERS; ++slot) {
        StringList_t players;
        listPlayers(slot, players);
        for (uint32_t i = 0; i < players.size(); ++i) {
            if (players[i] == player) {
                slots += slot;
            }
        }
    }
    return slots;
}

/** Get set of all slots in this game. */
GPlayerSet
Game::getGameSlots()
{
    GPlayerSet result;
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (isSlotInGame(i)) {
            result += i;
        }
    }
    return result;
}

/** Clear cache. */
void
Game::clearCache()
{
    game.hashKey("cache").remove();
}

/** Get game difficulty if it is not yet known. */
int32_t
Game::getDifficulty()
{
    DbHashIntField f = game.hashKey("cache").intField("difficulty");
    std::auto_ptr<IntValue> iv(f.getRaw());
    if (!iv.get()) {
        int value = computeGameRating(game.getClient(), *this);
        f.set(value);
        return value;
    } else {
        return toIntResult(iv.get());
    }
}

/** Mark this game broken. */
void
Game::markBroken(string_t message)
{
    game.stringKey("crashmessage").set(message);
    root.intSetKey("broken").add(gameId);
}

/** Get schedule subtree. */
DbSubtree
Game::getSchedule()
{
    return DbSubtree(game.subtree("schedule"));
}

/** Get game subtree. Use with care. */
DbSubtree
Game::getGame()
{
    return game;
}

/** Check whether a user was on a game. */
bool
Game::isUserOnGame(string_t user)
{
    return game.hashKey("users").hasField(user);
}

/** Check whether a user is on the game as primary user. */
bool
Game::isUserOnGameAsPrimary(string_t user)
{
    // Quick check first
    if (!isUserOnGame(user)) {
        return false;
    }

    // Check slots
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        Slot s = getSlot(i);
        if (s.slotStatus().get() != 0 && s.players()[0] == user) {
            return true;
        }
    }

    return false;
}

/** Check whether ranking is disabled in this game. */
bool
Game::isRankingDisabled()
{
    return getConfigInt("rankDisable") != 0 || getType() == "test";
}

/** Check whether joining as multiple races is allowed. */
bool
Game::isMultiJoinAllowed()
{
    return getConfigInt("joinMulti") != 0 || getType() == "test";
}

/** Describe this game. */
IntValue*
Game::describe(bool verbose, string_t forUser)
{
    /* @type HostGameInfo
       Information about a game.
       Game information can be provided in a "normal" and "verbose" format.
       Items marked ** are only included in verbose reports.
       @key id:GID                       (Game Id)
       @key state:HostGameState          (Game state)
       @key type:HostGameType            (Game type)
       @key name:Str                     (Game name, {game:$GID:name})
       @key description:Str              (**Game description)
       @key difficulty:Int               (Difficulty)
       @key currentSchedule:HostSchedule (current schedule)
       @key slots:StrList                (**Slot states. For each slot, one of "occupied", "open", "self", or "dead")
       @key turns:IntList                (**Turn states. For each slot, a {@type HostTurnStatus} value)
       @key joinable:Int                 (**1 if user can join this game)
       @key scores:IntList               (**Scores, if game is running or finished)
       @key scoreName:Str                (**Score Id)
       @key scoreDescription:Str         (**Score description)
       @key host:Str                     (Host Id, {game:$GID:settings}->host)
       @key hostDescription:Str          (Description for that host)
       @key shiplist:Str                 (Shiplist Id, {game:$GID:settings}->shiplist)
       @key shiplistDescription:Str      (Description for that shiplist)
       @key master:Str                   (**Master Id, {game:$GID:settings}->master)
       @key masterDescription:Str        (**Description for that master)
       @key turn:Int                     (turn number)
       @key lastHostTime:Time            (last host time)
       @key nextHostTime:Time            (next host time, if known)
       @key forum:FID                    (**associated forum, if any) */
    Ptr<IntArrayData> a = new IntArrayData();

    int32_t turnNr = game.hashKey("settings").getIntField("turn");
    string_t state = getState();

    // Id
    a->content.pushNew(makeStringValue("id"));
    a->content.pushNew(makeIntValue(gameId));

    // State
    a->content.pushNew(makeStringValue("state"));
    a->content.pushNew(makeStringValue(state));

    // Type
    a->content.pushNew(makeStringValue("type"));
    a->content.pushNew(makeStringValue(getType()));

    // Name
    a->content.pushNew(makeStringValue("name"));
    a->content.pushNew(makeStringValue(getName()));

    // Description
    if (verbose) {
        a->content.pushNew(makeStringValue("description"));
        a->content.pushNew(makeStringValue(game.hashKey("settings").getStringField("description")));
    }

    // Difficulty
    a->content.pushNew(makeStringValue("difficulty"));
    a->content.pushNew(makeIntValue(getDifficulty()));

    // Schedule
    DbSubtree schedule = game.subtree("schedule");
    string_t currentSchedule = schedule.stringListKey("list")[0];
    if (!currentSchedule.empty()) {
        a->content.pushNew(makeStringValue("currentSchedule"));
        a->content.pushNew(packSchedule(schedule.hashKey(currentSchedule)));
    }

    // Slot states
    if (verbose) {
        Ptr<IntArrayData> b = new IntArrayData();
        Ptr<IntArrayData> c = new IntArrayData();
        bool onGameAsPrimary = false;
        for (int i = 1; i <= 11; ++i) {
            Slot slot(getSlot(i));
            int turnState;
            if (slot.slotStatus().get() != 0) {
                StringList_t players;
                slot.players().getAll(players);
                int size = players.size();
                if (size == 0) {
                    b->content.pushNew(makeStringValue("open"));
                    turnState = 0;
                } else if (std::find(players.begin(), players.end(), forUser) != players.end()) {
                    b->content.pushNew(makeStringValue("self"));
                    turnState = slot.turnStatus().get();
                    if (players[0] == forUser) {
                        onGameAsPrimary = true;
                    }
                } else {
                    b->content.pushNew(makeStringValue("occupied"));
                    turnState = slot.turnStatus().get();
                    if (!forUser.empty()) {
                        turnState &= TurnStateMask;
                        if (turnState == TurnYellow || turnState == TurnGreen) {
                            turnState = TurnGreen;
                        } else if (turnState == TurnDead) {
                            turnState = TurnDead;
                        } else {
                            turnState = TurnMissing;
                        }
                    }
                }
            } else {
                b->content.pushNew(makeStringValue("dead"));
                turnState = 0;
            }
            c->content.pushNew(makeIntValue(turnState));
        }
        b->addDimension(b->content.size());
        c->addDimension(c->content.size());
        a->content.pushNew(makeStringValue("slots"));
        a->content.pushNew(new IntArray(b));
        a->content.pushNew(makeStringValue("turns"));
        a->content.pushNew(new IntArray(c));
        a->content.pushNew(makeStringValue("joinable"));
        a->content.pushNew(makeIntValue(!onGameAsPrimary || isMultiJoinAllowed()));
    }

    // Scores
    if (verbose && (state == "running" || state == "finished")) {
        string_t scoredesc = game.hashKey("scores").getStringField("score");
        string_t scorename;
        if (scoredesc.empty()) {
            scoredesc = "Classic Score";
            scorename = "timscore";
        } else {
            scorename = "score";
        }
        string_t scores = game.subtree("turn").subtree(turnNr).hashKey("scores").getStringField(scorename);
        if (scores.size() == 4*NUM_PLAYERS) {
            Ptr<IntArrayData> b = new IntArrayData();
            for (int i = 0; i < NUM_PLAYERS; ++i) {
                b->content.pushNew(makeIntValue(getInt32(scores.data() + 4*i)));
            }
            b->addDimension(b->content.size());
            a->content.pushNew(makeStringValue("scores"));
            a->content.pushNew(new IntArray(b));
            a->content.pushNew(makeStringValue("scoreName"));
            a->content.pushNew(makeStringValue(scorename));
            a->content.pushNew(makeStringValue("scoreDescription"));
            a->content.pushNew(makeStringValue(scoredesc));
        }
    }

    // Host
    string_t host = game.hashKey("settings").getStringField("host");
    a->content.pushNew(makeStringValue("host"));
    a->content.pushNew(makeStringValue(host));

    a->content.pushNew(makeStringValue("hostDescription"));
    a->content.pushNew(makeStringValue(DbSubtree(database_connection, HOST_PROGRAM_ROOT).hashKey("prog:" + host).getStringField("description")));

    // Ship list
    string_t shiplist = game.hashKey("settings").getStringField("shiplist");
    a->content.pushNew(makeStringValue("shiplist"));
    a->content.pushNew(makeStringValue(shiplist));

    a->content.pushNew(makeStringValue("shiplistDescription"));
    a->content.pushNew(makeStringValue(DbSubtree(database_connection, SHIPLIST_PROGRAM_ROOT).hashKey("prog:" + shiplist).getStringField("description")));

    // Master
    if (verbose) {
        string_t master = game.hashKey("settings").getStringField("master");
        a->content.pushNew(makeStringValue("master"));
        a->content.pushNew(makeStringValue(master));

        a->content.pushNew(makeStringValue("masterDescription"));
        a->content.pushNew(makeStringValue(DbSubtree(database_connection, MASTER_PROGRAM_ROOT).hashKey("prog:" + master).getStringField("description")));
    }

    // Turn
    a->content.pushNew(makeStringValue("turn"));
    a->content.pushNew(makeIntValue(turnNr));

    // Host times
    int32_t lastHostTime = getConfigInt("lastHostTime");
    a->content.pushNew(makeStringValue("lastHostTime"));
    a->content.pushNew(makeIntValue(getUserTimeFromTime(lastHostTime)));

    string_t nextHostTimeStr = getConfig("nextHostTime");
    int32_t nextHostTime;
    if (!nextHostTimeStr.empty() && stringValue(nextHostTimeStr, nextHostTime)) {
        a->content.pushNew(makeStringValue("nextHostTime"));
        a->content.pushNew(makeIntValue(getUserTimeFromTime(nextHostTime)));
    }

    // Forum
    if (verbose) {
        a->content.pushNew(makeStringValue("forum"));
        a->content.pushNew(makeIntValue(getConfigInt("forum")));
    }

    a->addDimension(a->content.size());

    return new IntArray(a);
}

/** Describe a slot. */
IntValue*
Game::describeSlot(int32_t slot, string_t forUser)
{
    /* @type HostPlayerInfo
       Information about a player in a game.
       @key long:Str       Long race name
       @key short:Str      Short race name
       @key adj:Str        Race name adjective
       @key users:StrList  {@type UID|User Ids} of all users in this slot.
                           The primary player is on the first slot,
                           replacements on following slots.
       @key editable:Int   Number of editable users (number of users you can call {PLAYERRESIGN} for)
                           at the end of the %users list.
       @key joinable:Int   1=you can join this slot */
    Ptr<IntArrayData> a = new IntArrayData();

    a->content.pushNew(makeStringValue("long"));
    a->content.pushNew(makeStringValue(raceNames.getFullName(slot)));

    a->content.pushNew(makeStringValue("short"));
    a->content.pushNew(makeStringValue(raceNames.getShortName(slot)));

    a->content.pushNew(makeStringValue("adj"));
    a->content.pushNew(makeStringValue(raceNames.getAdjName(slot)));

    bool counting = (forUser.empty() || forUser == getOwner());
    int32_t numEditable = 0;
    bool occupied;

    {
        Ptr<IntArrayData> au = new IntArrayData();
        StringList_t users;
        listPlayers(slot, users);
        for (uint32_t i = 0; i < users.size(); ++i) {
            if (forUser == users[i]) {
                counting = true;
            }
            if (counting) {
                ++numEditable;
            }
            au->content.pushNew(makeStringValue(users[i]));
        }
        au->addDimension(au->content.size());
        a->content.pushNew(makeStringValue("users"));
        a->content.pushNew(new IntArray(au));
        occupied = users.size() > 0;
    }

    a->content.pushNew(makeStringValue("editable"));
    a->content.pushNew(makeIntValue(numEditable));

    a->content.pushNew(makeStringValue("joinable"));
    a->content.pushNew(makeIntValue(!occupied && (!isUserOnGameAsPrimary(forUser) || isMultiJoinAllowed())));

    a->addDimension(a->content.size());
    return new IntArray(a);
}

bool
Game::hasPermission(string_t user, PermissionLevel level)
{
    // xref talk/htmlout.cc, renderGameLink
    // Admin has all permissions
    if (user.empty()) {
        return true;
    }

    switch (level) {
     case ReadPermission:
        // Everyone can read joining/running/finished x unlisted/public
        {
            string_t state = getState();
            if (state != "joining" && state != "running" && state != "finished") {
                return getOwner() == user;
            }

            string_t type = getType();
            return type == "unlisted"
                || type == "public"
                || getOwner() == user
                || isUserOnGame(user);
        }

     case ConfigPermission:
     case AdminPermission:
        // Only owner has these permissions
        return getOwner() == user;
    }
    return false;

// - admin         --> admin permissions (=no USER command)
// - none          --> no permissions needed, always available
// - game:read     --> everyone for joining/running/finished x unlisted/public, owner+players for j/r/f private/test
// - game:admin    --> ?
// - game:config   --> owner
}

static bool
tryLoadRaceNames(GRaceNameList& raceNames, string_t dir)
{
    try {
        MemoryStream content(file_connection.callString(DbRequest().withString("GET").withString(dir + "/race.nm")));
        raceNames.load(content);
        return true;
    }
    catch (std::exception& e) {
        return false;
    }
}

/** Load race names. This calls the file server. */
void
Game::loadRaceNames(DbClient& dbConnection)
{
    // Read race.nm file. Locations to try:
    // - game directory
    // - shiplist directory
    // - master directory
    // - host directory
    // - defaults
    // Note that we need admin permissions to read all these files.
    ReconnectingGuard g(file_rc);

    // FIXME: character set? This forces latin-1, which is the default.
    file_connection.callVoid(DbRequest().withString("USER").withString(""));
    if (tryLoadRaceNames(raceNames, game.stringKey("dir").get() + "/data")
        || tryLoadRaceNames(raceNames,
                            DbSubtree(dbConnection, SHIPLIST_PROGRAM_ROOT).subtree("prog")
                            .hashKey(getConfig("shiplist")).getStringField("path"))
        || tryLoadRaceNames(raceNames,
                            DbSubtree(dbConnection, MASTER_PROGRAM_ROOT).subtree("prog")
                            .hashKey(getConfig("master")).getStringField("path"))
        || tryLoadRaceNames(raceNames,
                            DbSubtree(dbConnection, HOST_PROGRAM_ROOT).subtree("prog")
                            .hashKey(getConfig("host")).getStringField("path"))
        || tryLoadRaceNames(raceNames, "defaults"))
    {
        // ok
    } else {
        // not ok. We could generate some defaults, but that would fail
        // when the game is ultimately hosted.
    }
}

DbHashIntField
Game::configChanged()
{
    return game.hashKey("settings").intField("configChanged");
}

DbHashIntField
Game::scheduleChanged()
{
    return game.hashKey("settings").intField("scheduleChanged");
}

DbHashIntField
Game::endChanged()
{
    return game.hashKey("settings").intField("endChanged");
}
