/**
  *  \file planetscentral/mailout/mailout.cc
  *  \brief Mailout: Main Program
  */

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include "planetscentral/mailout/mailout.h"
#include "planetscentral/dbclient/dbclient.h"
#include "planetscentral/osutil.h"
#include "planetscentral/dbclient/reconnect.h"
#include "util/misc.h"
#include "server/config.h"
#include "server/net.h"
#include "int/args.h"
#include "int/array.h"
#include "planetscentral/mailout/db.h"
#include "planetscentral/mailout/transmitter.h"
#include "planetscentral/net/smtp.h"

static string_t arg_listen_host = "127.0.0.1";
static string_t arg_listen_port = "21212";
static int32_t arg_threads = 5;

Mutex output_mutex;
DbReadyClient database_connection;
ReconnectingConnection database_rc(database_connection, output_mutex, "127.0.0.1", "6379");
Transmitter transmitter;

static void
processConfig(string_t key, string_t value)
{
    if (Smtp::checkConfig(key, value)) {
        // ok
    } else if (Transmitter::checkConfig(key, value)) {
        // ok
    } else if (key == "MAILOUT.HOST") {
        /* @q Mailout.Host:Str (Config)
           Listen address for the Mailout service. */
        arg_listen_host = value;
    } else if (key == "MAILOUT.PORT") {
        /* @q Mailout.Port:Int (Config)
           Port number for the Mailout service. */
        arg_listen_port = value;
    } else if (key == "MAILOUT.THREADS") {
        /* @q Mailout.Threads:Int (Config)
           Number of threads (=maximum number of parallel connections). */
        stringValue(value, arg_threads);
    } else if (key == "MAILOUT.TEMPLATEDIR") {
        /* @q Mailout.TemplateDir:Str (Config)
           Directory containing template files for outgoing mails. */
        transmitter.setTemplateDir(value);
    } else if (key == "REDIS.HOST") {
        database_rc.setHost(value);
    } else if (key == "REDIS.PORT") {
        database_rc.setPort(value);
    }
}

static bool
isPrintable(const string_t& s)
{
    for (string_t::size_type i = 0; i < s.size(); ++i)
        if (!isalnum(uint8_t(s[i])) && strchr("/._-:@,=", s[i]) == 0)
            return false;
    return true;
}

/*************************** MailoutConnection ***************************/

MailoutConnection::MailoutConnection()
{ }

void
MailoutConnection::handleRequest()
{
    // Read request
    std::auto_ptr<IntValue> req(readObject(true));
    IntArray* a = dynamic_cast<IntArray*>(req.get());
    if (!a || a->getData()->getDimension(0) == 0) {
        sendError("400 Syntax error");
        return;
    }
    const IntDataSegment& seg = a->getData()->content;

    // Log request
    {
        MutexGuard g(output_mutex);
        printf("[cmd]");
        for (uint32_t i = 0; i < seg.size(); ++i) {
            IntStringValue* sv = dynamic_cast<IntStringValue*>(seg[i]);
            if (sv != 0) {
                if (sv->getValue().size() < 100 && isPrintable(sv->getValue())) {
                    printf(" %s", sv->getValue().c_str());
                } else {
                    printf(" ...");
                }
            } else {
                printf(" %s", IntValue::toStringOf(seg[i], false).c_str());
            }
        }
        printf("\n");
    }

    string_t cmd = strUCase(IntValue::toStringOf(seg[0], false));
    IntArgBlock args(seg, 1, seg.size()-1);
    handleCommand(cmd, args);
}

void
MailoutConnection::handleCommand(const string_t& cmd, IntArgBlock& args)
{
    // Command dispatcher
    if (cmd == "PING") {
        /* @q PING (Mailout Command)
           Alive test.
           @retval Str "PONG". */
        sendSuccess("PONG");
    } else if (cmd == "HELP") {
        /* @q HELP (Mailout Command)
           @retval Str Help page. */
        IntStringValue sv("Mailout\n"
                          " PING\n"
                          " CONFIRM mail key [info]\n"
                          " CANCEL uid\n"
                          " REQUEST user\n"
                          " RUNQUEUE\n"
                          "Send mail:\n"
                          " MAIL tpl [uid]\n"
                          " PARAM name value\n"
                          " ATTACH url\n"
                          " SEND user|mail...\n");
        sendObject(&sv);
    } else if (cmd == "MAIL") {
        doMailNew(args);
    } else if (cmd == "PARAM") {
        doMailParam(args);
    } else if (cmd == "ATTACH") {
        doMailAttach(args);
    } else if (cmd == "SEND") {
        doMailSend(args);
    } else if (cmd == "CONFIRM") {
        doConfirm(args);
    } else if (cmd == "REQUEST") {
        doRequest(args);
    } else if (cmd == "CANCEL") {
        doCancel(args);
    } else if (cmd == "RUNQUEUE") {
        /* @q RUNQUEUE (Mailout Command)
           Trigger sending of queued mail. */
        args.checkArgumentCount(0);
        transmitter.runQueue();
        sendSuccess("OK");
    } else {
        sendError("400 Unknown command");
    }
}

/* @q MAIL tpl:Str, [uniq:Str] (Mailout Command)
   Start a new mail.
   If a %uniq is specified, a previous message with this value will become obsolete.
   @err 406 Sequence error (an unfinished transaction is active)
   @uses mqueue:uniqid, mqueue:msg:id, mqueue:preparing */
void
MailoutConnection::doMailNew(IntArgBlock& args)
{
    // Check
    args.checkArgumentCount(1, 2);
    if (msg.get() != 0) {
        sendError("406 Sequence error");
        return;
    }

    // Work
    ReconnectingGuard g(database_rc);
    DbSubtree root(database_connection, MAIL_ROOT);
    msg = allocateMessage(root);
    msg->templateName().set(toStringResult(args.getNext()));
    if (args.getNumArgs() > 0) {
        msg->uniqueId().set(toStringResult(args.getNext()));
    }
    sendSuccess("OK");
}

/* @q PARAM name:Str, val:Str (Mailout Command)
   Define a template parameter for the current mail.
   @err 406 Sequence error (no preceding {MAIL} command)
   @uses mqueue:msg:$ID:args */
void
MailoutConnection::doMailParam(IntArgBlock& args)
{
    // Check
    args.checkArgumentCount(2);
    if (msg.get() == 0) {
        sendError("406 Sequence error");
        return;
    }
    string_t name = toStringResult(args.getNext());
    string_t value = toStringResult(args.getNext());

    // Work
    ReconnectingGuard g(database_rc);
    msg->arguments().setField(name, value);
    sendSuccess("OK");
}

/* @q ATTACH url:Str (Mailout Command)
   Create attachment for the current mail.
   See {mqueue:msg:$ID:attach} for a format description.

   Note that this will only remember the attachment URL, not actually resolve it.
   Resolution will happen at an unspecified later time.
   @err 406 Sequence error (no preceding {MAIL} command)
   @uses mqueue:msg:$ID:attach */
void
MailoutConnection::doMailAttach(IntArgBlock& args)
{
    // Check
    args.checkArgumentCount(1);
    if (msg.get() == 0) {
        sendError("406 Sequence error");
        return;
    }

    // Work
    ReconnectingGuard g(database_rc);
    msg->attachments().pushBack(toStringResult(args.getNext()));
    sendSuccess("OK");
}

/* @q SEND addr:Str... (Mailout Command)
   Send current mail.
   Any number of receivers can be specified.
   See {mqueue:msg:$ID:to} for valid receiver formats.

   This command will eventually send the message.
   Message transmission happens asynchronously.

   @err 406 Sequence error (no preceding {MAIL} command)
   @uses mqueue:preparing, mqueue:sending, mqueue:msg:$ID:to */
void
MailoutConnection::doMailSend(IntArgBlock& args)
{
    // Check
    if (msg.get() == 0) {
        sendError("406 Sequence error");
        return;
    }

    // Work
    ReconnectingGuard g(database_rc);
    while (args.getNumArgs() > 0) {
        msg->receivers().add(toStringResult(args.getNext()));
    }
    {
        MutexGuard g(output_mutex);
        printf("[cmd] [msg%d] queued\n", int(msg->getId()));
    }
    msg->send();
    transmitter.send(msg->getId());
    msg.reset();
    sendSuccess("OK, queued");
}

/* @q CANCEL uniq:Str (Mailout Command)
   Cancels the mail with the given %uniq Id. */
void
MailoutConnection::doCancel(IntArgBlock& args)
{
    // Check
    args.checkArgumentCount(1);

    // Work
    ReconnectingGuard g(database_rc);
    DbSubtree root(database_connection, MAIL_ROOT);
    root.hashKey("uniqid").removeField(toStringResult(args.getNext()));
    sendSuccess("OK");
}

/* @q CONFIRM addr:Str, key:Str, [info:Str] (Mailout Command)
   Confirm email address using a confirmation key.
   This registers the email address as valid and starts sending mail to it.
   @err 401 Authentication error (key is not valid) */
void
MailoutConnection::doConfirm(IntArgBlock& args)
{
    // Check
    args.checkArgumentCount(2, 3);
    string_t mail = toStringResult(args.getNext());
    string_t key = toStringResult(args.getNext());
    string_t info = toStringResult(args.getNext());
    if (confirmMail(mail, key, info)) {
        transmitter.notifyAddress(mail);
        sendSuccess("OK");
    } else {
        sendError("401 Authentication error");
    }
}

/* @q REQUEST user:UID (Mailout Command)
   Request confirmation for a user.
   If the user has an unconfirmed, unrequested email address, sends the confirmation request.
   @err 400 Unable to request this (request cannot be sent, for example, user has no email address) */
void
MailoutConnection::doRequest(IntArgBlock& args)
{
    // Check
    args.checkArgumentCount(1);
    string_t userId = toStringResult(args.getNext());
    if (transmitter.maybeRequestConfirmation(userId)) {
        sendSuccess("OK");
    } else {
        sendError("400 Unable to request this");
    }
}


static int server_socket;

static void*
serverLoop(void*)
{
    while (1) {
        MailoutConnection conn;
        if (!conn.accept(server_socket)) {
            perror("accept");
            sleep(10);
        } else {
            try {
                while (!conn.isAtEnd()) {
                    try {
                        conn.handleRequest();
                    }
                    catch (std::exception& e) {
                        {
                            MutexGuard g(output_mutex);
                            printf("[cmd] [exception] %s\n", e.what());
                        }
                        conn.sendError(e.what());
                    }
                    conn.flush();
                }
            }
            catch (std::exception& e) {
                printf("c2mailout: %s\n", e.what());
            }
            conn.close();
        }
    }
}

int main(int /*argc*/, char** argv)
{
    /* Args */
    bool opt_notx = false;
    while (const char* p = *++argv) {
        if (p[0] == '-' && p[1] == '-') {
            ++p;
        }
        if (strcmp(p, "-notx") == 0) {
            opt_notx = true;
        } else {
            printf("c2mailout: invalid parameter.\n\n"
                   "Usage: c2mailout [-notx]\n");
            return 1;
        }
    }

    /* Unbuffer stdout */
    setvbuf(stdout, 0, _IOLBF, 0);

    /* Configuration */
    loadConfiguration(processConfig);

    /* Connect to database */
    try {
        database_rc.connectRetry();
    }
    catch (std::exception& e) {
        fprintf(stderr, "unable to connect to servers: %s\n", e.what());
        return 1;
    }

    /* Start transmitter */
    if (!opt_notx) {
        try {
            transmitter.start();
        }
        catch (std::exception& e) {
            fprintf(stderr, "unable to start transmitter: %s\n", e.what());
            return 1;
        }
    }

    /* Listen */
    server_socket = Bind().withAddress(arg_listen_host, arg_listen_port);
    if (server_socket < 0) {
        fprintf(stderr, "[main] unable to bind socket\n");
        return 1;
    }

    /* OK, main loop */
    printf("[main] listening...\n");
    signal(SIGPIPE, SIG_IGN);
    fcntl(server_socket, F_SETFD, FD_CLOEXEC);
    listen(server_socket, 10);

    /* Create some threads. Create one less than configured; the main
       thread will also serve requests. */
    for (int32_t i = 1; i < arg_threads; ++i) {
        pthread_t handle;
        if (pthread_create(&handle, 0, serverLoop, 0) < 0) {
            perror("[main] unable to create thread");
            return 1;
        }
    }
    serverLoop(0);
    close(server_socket);
}
