/**
  *  \file planetscentral/talk/cmdpm.cc
  *  \brief Talk: PM commands
  */

#include <set>
#include <stdexcept>
#include "planetscentral/talk/cmdpm.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/talk/talkconnection.h"
#include "planetscentral/talk/user.h"
#include "planetscentral/talk/userfolder.h"
#include "planetscentral/talk/talk.h"
#include "planetscentral/talk/userpm.h"
#include "planetscentral/talk/forum.h"      // getTime
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/notify.h"
#include "util/misc.h"
#include "game/struct.h"

/** Parse a single receiver.
    \param in Receiver
    \param out Individual users will be reported here */
static void
parseReceiver(const string_t& in, std::set<string_t>& out)
{
    if (in.size() > 2 && in.compare(0, 2, "u:", 2) == 0) {
        out.insert(string_t(in, 2, in.size()-2));
    } else if (in.size() > 2 && in.compare(0, 2, "g:", 2) == 0) {
        string_t::size_type divi = in.find(':', 2);
        if (divi == string_t::npos) {
            // Parse
            int32_t gameId;
            if (!stringValue(string_t(in, 2), gameId)) {
                throw std::runtime_error("400 Receiver syntax error");
            }

            // Validate
            DbSubtree g(database_connection, GAME_ROOT);
            if (!g.intSetKey("all").contains(gameId) || g.subtree(gameId).stringKey("state").get() == "deleted") {
                throw std::runtime_error("400 Receiver not found");
            }

            // Get
            StringList_t result;
            g.subtree(gameId).hashKey("users").getAll(result);
            for (StringList_t::size_type i = 0; i < result.size(); i += 2) {
                if (result[i+1] != "0") {
                    out.insert(result[i]);
                }
            }
        } else {
            // Parse
            int32_t gameId, slotId;
            if (!stringValue(string_t(in, 2, divi-2), gameId)
                || !stringValue(string_t(in, divi+1), slotId)
                || slotId <= 0 || slotId > NUM_PLAYERS)
            {
                throw std::runtime_error("400 Receiver syntax error");
            }

            // Validate
            DbSubtree g(database_connection, GAME_ROOT);
            if (!g.intSetKey("all").contains(gameId) || g.subtree(gameId).stringKey("state").get() == "deleted") {
                throw std::runtime_error("400 Receiver not found");
            }

            // Get
            StringList_t users;
            g.subtree(gameId).subtree("player").subtree(slotId).stringListKey("users").getAll(users);
            for (StringList_t::size_type i = 0; i < users.size(); ++i) {
                out.insert(users[i]);
            }
        }
    } else {
        throw std::runtime_error("400 Receiver syntax error");
    }
}

/** Parse a receiver list.
    \param in Receiver list as given in the command
    \param out Individual users will be reported here */
static void
parseReceivers(const string_t& in, std::set<string_t>& out)
{
    // This is a machine interface, so there's no need to normalize.
    // 'in' is expected to contain comma-separated receivers, and that's it.
    string_t::size_type a = 0;
    string_t::size_type b;
    while ((b = in.find(',', a)) != string_t::npos) {
        parseReceiver(string_t(in, a, b-a), out);
        a = b+1;
    }
    parseReceiver(string_t(in, a, in.size()-a), out);
}

/* @q PMNEW to:TalkAddr subject:Str text:TalkText [PARENT parent:PMID] (Talk Command)
   Send a PM.

   New message attributes:
   - author: current user
   - to, subject, text, parent: as given
   - time: current time
   The new message is stored in the sender's outbox (folder 2),
   and in all receivers' inboxes (folder 1).

   Permissions: user context required

   @err 400 Receiver syntax error
   @err 400 Receiver not found
   @err 412 No receivers (receiver list expands to no users at all, e.g. when sending to an empty game)

   @retval PMID Id of new message
   @uses pm:id, pm:$PMID:header, pm:$PMID:text, user:$UID:pm:folder:$UFID:messages
   @uses MAIL (Mailout Command) */
void
doPMNew(TalkConnection& conn, IntArgBlock& args)
{
    // Parse
    args.checkArgumentCountAtLeast(3);
    conn.checkUser();
    string_t to      = toStringResult(args.getNext());
    string_t subject = toStringResult(args.getNext());
    string_t text    = toStringResult(args.getNext());
    int32_t parent = 0;
    while (args.getNumArgs() > 0) {
        string_t key = strUCase(toStringResult(args.getNext()));
        if (key == "PARENT") {
            args.checkArgumentCountAtLeast(1);
            parent = toIntResult(args.getNext());
        } else {
            conn.sendError("400 Invalid parameter");
            return;
        }
    }

    // Execute
    ReconnectingGuard g(database_rc);
    std::set<string_t> recv;
    parseReceivers(to, recv);
    if (recv.empty()) {
        conn.sendError("412 No receivers");
        return;
    }

    // Create the message
    int32_t pmid = UserPM::allocatePM(database_connection);
    UserPM pm(database_connection, pmid);

    const string_t sender = conn.getUser();
    pm.author().set(sender);
    pm.receivers().set(to);
    pm.subject().set(subject);
    pm.time().set(getTime());
    pm.text().set(text);
    pm.flags(sender).set(PMStateRead);
    if (parent != 0) {
        pm.parentMessageId().set(parent);
    }

    // Distribute the message
    StringList_t notifyIndividual;
    StringList_t notifyGroup;
    {
        User u(database_connection, sender);
        if (UserFolder(u, PMSystemOutboxFolder).messages().add(pmid)) {
            pm.addReference();
        }
    }
    for (std::set<string_t>::iterator i = recv.begin(); i != recv.end(); ++i) {
        User u(database_connection, *i);
        UserFolder folder(u, PMSystemInboxFolder);
        if (folder.messages().add(pmid)) {
            pm.addReference();

            // Process notifications
            if (*i != sender) {
                string_t notify = u.getPMMailType();
                if (notify != "none") {
                    if (notify == "info") {
                        if (folder.unreadMessages().get() == 0) {
                            notifyGroup.push_back("user:" + *i);
                        }
                    } else {
                        notifyIndividual.push_back("user:" + *i);
                    }
                }
                folder.unreadMessages().set(true);
            }
        }
    }

    // Send notifications
    notifyPM(pm, notifyIndividual, notifyGroup);

    // Result is message Id. Might be useful.
    IntIntIntValue iv(pmid);
    conn.sendObject(&iv);
}

/* @q PMSTAT folder:UFID msg:PMID (Talk Command)
   Get information about one message.

   Permissions: user context required, accesses user's PMs

   @err 404 Not found (PM does not exist)
   @retval TalkMailInfo information
   @uses pm:$PMID:header, user:$UID:pm:folder:$UFID:messages
   @see PMMSTAT */
void
doPMStat(TalkConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCount(2);
    conn.checkUser();
    int32_t ufid = toIntResult(args.getNext());
    int32_t pmid = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    User u(database_connection, conn.getUser());
    UserFolder folder(u, ufid);
    if (!folder.messages().contains(pmid)) {
        // No need to verify that the folder exists; if it does not exist,
        // it will report empty.
        conn.sendError("404 Not found");
    } else {
        // Report the message
        UserPM pm(database_connection, pmid);
        std::auto_ptr<IntValue> result(pm.describe(conn.getUser()));
        conn.sendObject(result.get());
    }
}

/* @q PMMSTAT folder:UFID msg:PMID... (Talk Command)
   Get information about multiple messages.

   Permissions: user context required, accesses user's PMs

   @retval TalkMailInfo[] information for each message; null if the message cannot be accessed
   @uses pm:$PMID:header, user:$UID:pm:folder:$UFID:messages
   @see PMSTAT */
void
doPMMStat(TalkConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCountAtLeast(1);
    conn.checkUser();
    int32_t ufid = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    User u(database_connection, conn.getUser());
    UserFolder folder(u, ufid);
    IntDataSegment result;
    while (args.getNumArgs() > 0) {
        int32_t pmid = toIntResult(args.getNext());
        if (!folder.messages().contains(pmid)) {
            result.pushNew(0);
        } else {
            UserPM pm(database_connection, pmid);
            result.pushNew(pm.describe(conn.getUser()));
        }
    }

    // Result
    conn.sendArray(result, result.size());
}

/* @q PMCP src:UFID dest:UFID msg:PMID... (Talk Command)
   Move messages between folders.
   This implements the semantic "make sure it's here AND there",
   i.e. does not fail if the target already exists.
   It is not an error if the message does not exist in the source folder.

   Permissions: user context required; accesses user's PMs.

   @retval Int number of messages successfully copied
   @uses user:$UID:pm:folder:$UFID:messages */
void
doPMCopy(TalkConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCountAtLeast(2);
    conn.checkUser();
    int32_t srcufid = toIntResult(args.getNext());
    int32_t dstufid = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    User u(database_connection, conn.getUser());
    UserFolder srcfolder(u, srcufid);
    UserFolder dstfolder(u, dstufid);

    // Verify that destination exists
    dstfolder.checkExistance();

    // Copy
    int32_t count = 0;
    while (args.getNumArgs() > 0) {
        int32_t pmid = toIntResult(args.getNext());
        if (srcfolder.messages().contains(pmid)) {
            if (dstfolder.messages().add(pmid)) {
                UserPM(database_connection, pmid).addReference();
            }
            ++count;
        }
    }

    IntIntIntValue rv(count);
    conn.sendObject(&rv);
}

/* @q PMMV src:UFID dest:UFID msg:PMID... (Talk Command)
   Move messages between folders.
   This implements the semantic "make sure it's not here but there",
   i.e. does not fail if the target already exists.
   It is not an error if the message does not exist in the source folder.

   Permissions: user context required; accesses user's PMs.

   @retval Int number of messages successfully moved
   @uses user:$UID:pm:folder:$UFID:messages */
void
doPMMove(TalkConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCountAtLeast(2);
    conn.checkUser();
    int32_t srcufid = toIntResult(args.getNext());
    int32_t dstufid = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    User u(database_connection, conn.getUser());
    UserFolder srcfolder(u, srcufid);
    UserFolder dstfolder(u, dstufid);

    // Verify that destination exists
    dstfolder.checkExistance();

    // Move
    int32_t count = 0;
    while (args.getNumArgs() > 0) {
        int32_t pmid = toIntResult(args.getNext());
        if (srcfolder.messages().remove(pmid)) {
            if (!dstfolder.messages().add(pmid)) {
                // I removed it from the source, but cannot add it to
                // the destination, so that's one lost reference.
                UserPM(database_connection, pmid).removeReference();
            }
            ++count;
        }
    }

    IntIntIntValue rv(count);
    conn.sendObject(&rv);
}

/* @q PMRM folder:UFID msg:PMID... (Talk Command)
   Delete messages.
   The specified messages are removed from the given folder.
   It is not an error to attempt to delete a message that does not exist.

   Permissions: user context required; accesses user's PMs.

   @retval Int number of messages successfully deleted
   @uses user:$UID:pm:folder:$UFID:messages */
void
doPMRemove(TalkConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCountAtLeast(1);
    conn.checkUser();
    int32_t ufid = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    User u(database_connection, conn.getUser());
    UserFolder folder(u, ufid);

    // Move
    int32_t count = 0;
    while (args.getNumArgs() > 0) {
        int32_t pmid = toIntResult(args.getNext());
        if (folder.messages().remove(pmid)) {
            UserPM(database_connection, pmid).removeReference();
            ++count;
        }
    }

    IntIntIntValue rv(count);
    conn.sendObject(&rv);
}

/* @q PMRENDER folder:UFID msg:PMID [renderOptions...] (Talk Command)
   Render a PM.
   The message is rendered using the current render options, see {RENDEROPTION}.
   You can temporarily override rendering options by specifying the new settings within the command.

   Permissions: user context required; accesses user's PMs.

   @err 404 Not found (Message does not exist in the specified folder)

   @retval Str rendered message
   @uses pm:$PMID:text
   @see PMMRENDER */
void
doPMRender(TalkConnection& conn, IntArgBlock& args)
{
    // Get parameters
    args.checkArgumentCountAtLeast(2);
    conn.checkUser();
    int32_t ufid = toIntResult(args.getNext());
    int32_t pmid = toIntResult(args.getNext());

    // Get options
    RenderState state(conn.getRenderState());
    while (args.getNumArgs() > 0) {
        string_t optUC = strUCase(toStringResult(args.getNext()));
        if (!state.handleOption(optUC, args)) {
            conn.sendError("400 Invalid option");
            return;
        }
    }

    // Do it
    ReconnectingGuard g(database_rc);
    User u(database_connection, conn.getUser());
    UserFolder folder(u, ufid);
    if (!folder.messages().contains(pmid)) {
        // No need to verify that the folder exists; if it does not exist,
        // it will report empty.
        conn.sendError("404 Not found");
    } else {
        // Report the message
        state.setMessageAuthor(UserPM(database_connection, pmid).author().get());
        IntStringValue sv(state.render(UserPM(database_connection, pmid).text().get(),
                                       database_connection));
        conn.sendObject(&sv);
    }
}

/* @q PMMRENDER folder:UFID msg:PMID... (Talk Command)
   Render PMs.
   Any number of messages can be specified,
   but all must be from the same folder because only one folder Id is given.

   The messages are rendered using the current render options, see {RENDEROPTION}.

   If one of the requested messages cannot be accessed,
   null is returned instead of a rendered message; no error is generated.

   Permissions: user context required; accesses user's PMs.

   @retval StrList rendered PMs
   @uses pm:$PMID:text
   @see PMRENDER */
void
doPMMRender(TalkConnection& conn, IntArgBlock& args)
{
    // Get parameters
    args.checkArgumentCountAtLeast(1);
    conn.checkUser();
    int32_t ufid = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    User u(database_connection, conn.getUser());
    UserFolder folder(u, ufid);
    RenderState state(conn.getRenderState());
    IntDataSegment result;
    while (args.getNumArgs() > 0) {
        int32_t pmid = toIntResult(args.getNext());
        if (!folder.messages().contains(pmid)) {
            result.pushNew(0);
        } else {
            // Render the message
            state.setMessageAuthor(UserPM(database_connection, pmid).author().get());
            result.pushNew(makeStringValue(state.render(UserPM(database_connection, pmid).text().get(),
                                                        database_connection)));
        }
    }

    // Result
    conn.sendArray(result, result.size());
}

/* @q PMFLAG folder:UFID clear:TalkFlag set:TalkFlag msg:PMID... (Talk Command)
   Change flags of a message.
   The %clear and %set parameters specify the bits to clear and set on all listed messages.
   %clear and %set should be mutually disjoint.

   Permissions: user context required; accesses user's PMs.

   @retval Int number of messages processed
   @uses pm:$PMID:header, user:$UID:pm:folder:$UFID:messages */
void
doPMFlag(TalkConnection& conn, IntArgBlock& args)
{
    // Get parameters
    args.checkArgumentCountAtLeast(3);
    conn.checkUser();
    int32_t ufid = toIntResult(args.getNext());
    int32_t clear = toIntResult(args.getNext());
    int32_t set = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    User u(database_connection, conn.getUser());
    UserFolder folder(u, ufid);
    int32_t result = 0;
    while (args.getNumArgs() > 0) {
        int32_t pmid = toIntResult(args.getNext());
        if (folder.messages().contains(pmid)) {
            UserPM msg(database_connection, pmid);
            msg.flags(conn.getUser()).set((msg.flags(conn.getUser()).get() & ~clear) | set);
            ++result;
        }
    }

    // Result
    IntIntIntValue iv(result);
    conn.sendObject(&iv);
}
