/**
  *  \file planetscentral/talk/notify.cc
  *  \brief Talk: Email notifications
  */

#include <algorithm>
#include "planetscentral/talk/notify.h"
#include "planetscentral/talk/message.h"
#include "planetscentral/talk/topic.h"
#include "planetscentral/talk/forum.h"
#include "planetscentral/talk/talk.h"
#include "planetscentral/talk/user.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/dbclient/mailclient.h"
#include "planetscentral/talk/url.h"
#include "planetscentral/talk/render.h"
#include "planetscentral/talk/userpm.h"
#include "planetscentral/talk/permission.h"

/** Notify a forum message. Sends mail to all users observing this topic or forum.
    \param msg Forum message
    \param topic Containing topic
    \param forum Containing forum */
void
notifyMessage(Message& msg, Topic& topic, Forum& forum)
{
    // Get sender
    const string_t author = msg.author().get();

    // -- Topic --
    // Get topic watchers
    StringList_t topicWatchers;
    topic.watchers().getAll(topicWatchers);
    std::sort(topicWatchers.begin(), topicWatchers.end());

    // Get post read permissions
    string_t readPermissions = topic.readPermissions().get();
    if (readPermissions.empty()) {
        readPermissions = forum.readPermissions().get();
    }

    // Find topic watchers that have not been notified and mark them.
    // The author of the posting will not be notified.
    StringList_t groupReceivers;
    StringList_t individualReceivers;
    for (StringList_t::size_type i = 0; i < topicWatchers.size(); ++i) {
        if (topicWatchers[i] != author && checkUserPermission(readPermissions, topicWatchers[i])) {
            User u(database_connection, topicWatchers[i]);
            if (u.isWatchIndividual()) {
                individualReceivers.push_back("user:" + topicWatchers[i]);
            } else if (u.notifiedTopics().add(topic.getId())) {
                groupReceivers.push_back("user:" + topicWatchers[i]);
            } else {
                // No notification for you
            }
        }
    }

    // Send message. We do not use a uniquifier here, because a future
    // message will have different receivers.
    if (!groupReceivers.empty()) {
        ReconnectingGuard g(mail_rc);
        mail_connection.startMail("talk-topic");
        mail_connection.addParameter("forum", forum.name().get());
        mail_connection.addParameter("subject", msg.subject().get());
        mail_connection.addParameter("posturl", makePostUrl(topic.getId(), topic.subject().get(), msg.getId()));
        mail_connection.sendTo(groupReceivers);
    }
    if (!individualReceivers.empty()) {
        RenderState rs;
        rs.setBaseUrl(www_prefix);
        rs.setFormat("mail");
        rs.setMessageId(msg.getId());
        
        ReconnectingGuard g(mail_rc);
        mail_connection.startMail("talk-topic-message");
        mail_connection.addParameter("forum", forum.name().get());
        mail_connection.addParameter("subject", msg.subject().get());
        mail_connection.addParameter("posturl", makePostUrl(topic.getId(), topic.subject().get(), msg.getId()));
        mail_connection.addParameter("message", rs.render(msg.text().get(), database_connection));
        mail_connection.addParameter("author", User(database_connection, author).getScreenName());
        mail_connection.sendTo(individualReceivers);
    }

    // -- Forum --
    // Get forum watchers
    StringList_t forumWatchers;
    forum.watchers().getAll(forumWatchers);
    std::sort(forumWatchers.begin(), forumWatchers.end());

    // Find forum watchers that have not been notified and mark them.
    // The author of the posting will not be notified.
    individualReceivers.clear();
    groupReceivers.clear();
    for (StringList_t::size_type i = 0; i < forumWatchers.size(); ++i) {
        if (forumWatchers[i] != author
            && !std::binary_search(topicWatchers.begin(), topicWatchers.end(), forumWatchers[i])
            && checkUserPermission(readPermissions, forumWatchers[i]))
        {
            User u(database_connection, forumWatchers[i]);
            if (u.isWatchIndividual()) {
                individualReceivers.push_back("user:" + forumWatchers[i]);
            } else if (u.notifiedForums().add(forum.getId())) {
                groupReceivers.push_back("user:" + forumWatchers[i]);
            } else {
                // No notification for you
            }
        }
    }

    // Send message. We do not use a uniquifier here, because a future
    // message will have different receivers.
    if (!groupReceivers.empty()) {
        ReconnectingGuard g(mail_rc);
        mail_connection.startMail("talk-forum");
        mail_connection.addParameter("forum", forum.name().get());
        mail_connection.addParameter("subject", msg.subject().get());
        mail_connection.addParameter("posturl", makePostUrl(topic.getId(), topic.subject().get(), msg.getId()));
        mail_connection.sendTo(groupReceivers);
    }
    if (!individualReceivers.empty()) {
        RenderState rs;
        rs.setBaseUrl(www_prefix);
        rs.setFormat("mail");
        rs.setMessageId(msg.getId());
        
        ReconnectingGuard g(mail_rc);
        mail_connection.startMail("talk-forum-message");
        mail_connection.addParameter("forum", forum.name().get());
        mail_connection.addParameter("subject", msg.subject().get());
        mail_connection.addParameter("posturl", makePostUrl(topic.getId(), topic.subject().get(), msg.getId()));
        mail_connection.addParameter("message", rs.render(msg.text().get(), database_connection));
        mail_connection.addParameter("author", User(database_connection, author).getScreenName());
        mail_connection.sendTo(individualReceivers);
    }
}

/** Notify a private message.
    \param msg Private message
    \param notifyIndividual Users that receive individual message notifications
    \param notifyGroup Users that receive group notifications */
void
notifyPM(UserPM& msg, const StringList_t& notifyIndividual, const StringList_t& notifyGroup)
{
    // Send notification with message
    if (!notifyIndividual.empty()) {
        RenderState rs;
        rs.setBaseUrl(www_prefix);
        rs.setFormat("mail");
        
        ReconnectingGuard g(mail_rc);
        mail_connection.startMail("talk-pm-message");
        mail_connection.addParameter("subject", msg.subject().get());
        mail_connection.addParameter("author", User(database_connection, msg.author().get()).getScreenName());
        mail_connection.addParameter("id", msg.getId());
        mail_connection.addParameter("message", rs.render(msg.text().get(), database_connection));
        mail_connection.sendTo(notifyIndividual);
    }

    // Send notification without message
    if (!notifyGroup.empty()) {
        ReconnectingGuard g(mail_rc);
        mail_connection.startMail("talk-pm");
        mail_connection.sendTo(notifyGroup);
    }
}
